import torch
import torch.utils.data

from .utils import memoize


class SafeDataset(torch.utils.data.Dataset):
    """A wrapper around a torch.utils.data.Dataset that allows dropping
    samples dynamically.
    """

    def __init__(self, dataset, eager_eval=False):
        """Creates a `SafeDataset` wrapper around `dataset`."""
        self.eager_eval = eager_eval
        # "Backup" of dataset's original __getitem__ function (since we will
        # be overwriting it)
        self._get_sample_original = dataset.__getitem__
        # _safe_get_item returns None instead of throwing an error when
        # dealing with a unsafe sample, and also builds an index of safe and
        # unsafe samples as it goes
        dataset.__class__.__getitem__ = self._safe_get_item
        self.dataset = dataset

        self.safe_samples_indices = []
        self.unsafe_samples_indices = []

        # If eager_eval is False, we can simply go ahead and build the index
        # by attempting to access every sample in self.dataset.
        if self.eager_eval is True:
            self._build_index()

    def _safe_get_item(self, idx):
        """Returns None instead of throwing an error when dealing with an
        unsafe sample, and also builds an index of safe and unsafe samples as
        and when they get accessed.
        """
        try:
            sample = self._get_sample_original(idx)
            if idx not in self.safe_samples_indices:
                self.safe_samples_indices.append(idx)
            return sample
        except IndexError: raise
        except Exception:
            if idx not in self.unsafe_samples_indices:
                self.unsafe_samples_indices.append(idx)
            return None

    def _build_index(self):
        for idx in range(len(self.dataset)):
            # The returned sample is deliberately discarded because
            # self.dataset[idx] is called only to classify every index
            # into either safe_samples_indices or unsafe_samples_indices.
            _ = self.dataset[idx]

    def _reset_index(self):
        """Resets the safe and unsafe samples indices."""
        self.safe_samples_indices = self.unsafe_samples_indices = []

    @property
    def is_index_built(self):
        """Returns True if all indices of the original dataset have been
        classified into safe_samples_indices or unsafe_samples_indices.
        """
        return len(self.dataset) == len(self.safe_samples_indices) + \
            len(self.unsafe_samples_indices)

    def __len__(self):
        """Returns the length of the original dataset.
        NOTE: This is different from the number of actually valid samples.
        """
        return len(self.dataset)

    def __iter__(self):
        return (self.dataset[i] for i in range(len(self.dataset)))

    @memoize
    def __getitem__(self, idx):
        if not self.is_index_built:
            raise RuntimeError('Index of safe samples is not yet built. ' +
                               'You can use this Dataset as an iterator. ' +
                               'However, you will not be able to index ' +
                               'arbitrary elements.')
        dataset_idx = self.safe_samples_indices[idx]
        return self.dataset[dataset_idx]



import torch
import torch.utils.data

from .dataset import SafeDataset


class SafeSampler(torch.utils.data.Sampler):
    """SafeSampler can be used both as a standard Sampler (over a Dataset),
    or as a wrapper around an existing `Sampler` instance. It allows you to
    drop unwanted samples while sampling.
    """

    def __init__(self, dataset, sampler=None, step_to_index_fn=None):
        """Create a `SafeSampler` instance that performs sampling over either
        another sampler object or directly over a dataset. `step_to_index_fn`
        will define the `SafeSampler` instance's behavior when it encounters
        an unsafe sample.

        When a `Sampler` object is passed to `sampler', the returned
        `SafeSampler` instance performs sampling over the list of indices
        sampled by `sampler`, acting as a wrapper over it. If `sampler` is
        `None`, the `SafeSampler` instance that is returned performs sampling
        directly over the `dataset` object.

        Arguments:
            dataset (SafeDataset): The dataset to be sampled.
            sampler (Sampler, optional): If `sampler` is `None`, the sampling
                is performed directly over the `dataset`, otherwise it's done
                over the list of indices returned by `sampler`'s `__iter__`
                method.
                If `sampler` takes a `Dataset` object as a parameter,
                `dataset` should ideally be the same as the one passed to
                `sampler`.
            step_to_index_fn (function, optional): Function that takes in 2
                arguments - (`num_valid_samples` and `num_samples_examined`),
                and returns the next index to be sampled. If None or not
                specified, the default function returns the
                `num_samples_examined` as the output.
        """
        assert isinstance(dataset, SafeDataset), \
            "Dataset must be a SafeDataset."
        self.dataset = dataset
        self.dataset.__class__.__getitem__ = self.dataset._get_sample_original

        self.sampler = sampler
        if sampler is not None:
            self.sampler_indices = list(iter(sampler))

        if step_to_index_fn is not None:
            self.step_to_index_fn = step_to_index_fn
        else:
            self.step_to_index_fn = lambda original, actual: actual

    def __iter__(self):
        """Return iterator over sampled indices."""
        self.num_valid_samples = self.num_samples_examined = 0
        return self

    def _get_next_index(self):
        """Helper function that calls `step_to_index_fn` and decides
        whether to sample directly from `dataset` or through `sampler`."""
        index = self.step_to_index_fn(
            self.num_valid_samples, self.num_samples_examined)
        if self.sampler is not None:
            index = self.sampler_indices[index]
        return index

    def __next__(self):
        """Returns next index to sample over `dataset`."""
        while True:
            try:
                index = self._get_next_index()
                self.num_samples_examined += 1
                if self.dataset[index] is not None:
                    self.num_valid_samples += 1
                    return index
            except IndexError:
                raise StopIteration

    # For Python2 compatibility
    next = __next__



from functools import partial


class memoize(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj(object):
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        print(cache)
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res



from setuptools import setup, find_packages


setup(name='nonechucks',
      version='0.1.1',
      url='https://github.com/msamogh/nonechucks',
      license='MIT',
      author='Amogh Mannekote',
      author_email='msamogh@gmail.com',
      description="""NoneChucks is a library for PyTorch that allows you to drop samples from your data and ensures that your Datasets and Transforms do not fail because of a few samples that are not parseable.""",
      packages=find_packages(),
      long_description=open('README.md').read(),
      zip_safe=False)
